import hashlib

import pandas
from pandas import DataFrame


def md5hash(pd: DataFrame) -> hashlib.md5:
    """
    API to make md5hash from pandas dataframe.

    Args:
        pd: pandas dataframe.
    Examples:
        >>> import pandas
        >>> pd = pandas.DataFrame({'a': [1, 10, 100, 1, 1, 100], 'b': ['a', 'b', 'c', 'd', 'e', 'f'],
        ...                       'idx': [0, 1, 2, 0, 1, 2]}).set_index('idx')
        >>> print(pd)
        ... # doctest: +NORMALIZE_WHITESPACE
               a  b
        idx
        0      1  a
        1     10  b
        2    100  c
        0      1  d
        1      1  e
        2    100  f
        >>> md5hash(pd)
        '80254916224f25f3701c2edc8afa8539'
    """
    hashes = pandas.util.hash_pandas_object(pd)
    m = hashlib.md5()
    for hash in hashes:
        m.update(hash.to_bytes(64, 'big'))
    return m.hexdigest()



import os
import shutil
from pathlib import Path
from unittest import TestCase

import dask.dataframe
from intake.source.csv import CSVSource

from framania.intakemania.extension import FramaniaExtendedIntakeCatalog, analysis


class TestFramaniaExtendedIntake(TestCase):
    def setUp(self) -> None:
        self.csvsource1 = CSVSource(urlpath='test/data/test-csv1.csv')
        self.csvsource2 = CSVSource(urlpath='test/data/test-csv2.csv')
        self.cpath = Path('test/temp/test-catalog.yaml')
        self.catalog = FramaniaExtendedIntakeCatalog(self.cpath)
        self.data_dir = Path('test/temp/data-dir')

    def tearDown(self) -> None:
        os.remove(str(self.cpath))
        shutil.rmtree(str(self.data_dir))

    def test_initialize(self):

        @analysis('raw_csv1', '1.0', catalog=self.catalog, sources=[], data_dir=self.data_dir)
        def load_raw_csv1(csv_source1):
            return csv_source1

        load_raw_csv1(self.csvsource1.to_dask())

        source = self.catalog['raw_csv1_1.0']
        intake_source = source.intake_source
        assert intake_source.metadata['extension'] == 'framania'
        assert intake_source.metadata['version'] == source.version == '1.0'
        assert intake_source.metadata['version_name'] == 'raw_csv1_1.0'
        assert intake_source.metadata['name'] == source.name == 'raw_csv1'
        assert intake_source.metadata['upstream'] == []
        assert intake_source.metadata['md5hash'] == source.md5hash == '12cd8e064eafadfc8f06af01d31fa8f5'

        assert source.upstream_sources == []

    def test_compare(self):
        @analysis('raw_csv1', '1.0', catalog=self.catalog, sources=[], data_dir=self.data_dir)
        def load_raw_csv1(csv_source1):
            return csv_source1

        raw_csv1_source, _ = load_raw_csv1(self.csvsource1.to_dask())

        @analysis('raw_csv2', '1.0', catalog=self.catalog, sources=[], data_dir=self.data_dir)
        def load_raw_csv2(csv_source2):
            return csv_source2

        raw_csv2_source, _ = load_raw_csv2(self.csvsource2.to_dask())

        assert raw_csv1_source.md5hash != raw_csv2_source.md5hash

        source = self.catalog['raw_csv2_1.0']
        intake_source = source.intake_source
        assert intake_source.metadata['extension'] == 'framania'
        assert intake_source.metadata['version'] == source.version == '1.0'
        assert intake_source.metadata['version_name'] == 'raw_csv2_1.0'
        assert intake_source.metadata['name'] == source.name == 'raw_csv2'
        assert intake_source.metadata['upstream'] == []
        assert intake_source.metadata['md5hash'] == source.md5hash == 'f4466ed4b50c808bfeb64681b4f33dda'

        assert source.upstream_sources == []

    def test_dag(self):
        @analysis('raw_csv1', '1.0', catalog=self.catalog, sources=[], data_dir=self.data_dir)
        def load_raw_csv1(csv_source1):
            return csv_source1

        raw_csv1_source, _ = load_raw_csv1(self.csvsource1.to_dask())

        @analysis('raw_csv2', '1.0', catalog=self.catalog, sources=[], data_dir=self.data_dir)
        def load_raw_csv2(csv_source2):
            return csv_source2

        raw_csv2_source, _ = load_raw_csv2(self.csvsource2.to_dask())

        @analysis('transform_csv1', '1.0', catalog=self.catalog, sources=[raw_csv1_source], data_dir=self.data_dir)
        def transform_csv1(raw_csv1):
            raw_csv1['a'] *= 10
            return raw_csv1

        @analysis('transform_csv1_2', '1.0', catalog=self.catalog, sources=[raw_csv1_source, raw_csv2_source],
                  data_dir=self.data_dir)
        def transform_csv1_2(raw_csv1, raw_csv2):
            raw_csv1_rename = raw_csv1.rename(columns={'a': 'a1'})
            raw_csv2_rename = raw_csv2.rename(columns={'a': 'a2'})
            return dask.dataframe.concat([raw_csv1_rename[['a1']], raw_csv2_rename[['a2']]], axis=1,
                                         interleave_partitions=True)
        transform_csv1()

        source = self.catalog['transform_csv1_1.0']
        intake_source = source.intake_source
        assert intake_source.metadata['extension'] == 'framania'
        assert intake_source.metadata['version'] == source.version == '1.0'
        assert intake_source.metadata['version_name'] == 'transform_csv1_1.0'
        assert intake_source.metadata['name'] == source.name == 'transform_csv1'
        assert intake_source.metadata['upstream'] == [{'version_name': 'raw_csv1_1.0',
                                                       'md5hash': '12cd8e064eafadfc8f06af01d31fa8f5'}]
        assert intake_source.metadata['md5hash'] == source.md5hash == 'f4466ed4b50c808bfeb64681b4f33dda'

        assert source.upstream_sources == [self.catalog['raw_csv1_1.0']]

        transform_csv1_2()

        source = self.catalog['transform_csv1_2_1.0']
        intake_source = source.intake_source
        assert intake_source.metadata['extension'] == 'framania'
        assert intake_source.metadata['version'] == source.version == '1.0'
        assert intake_source.metadata['version_name'] == 'transform_csv1_2_1.0'
        assert intake_source.metadata['name'] == source.name == 'transform_csv1_2'
        assert intake_source.metadata['upstream'] == [{'version_name': 'raw_csv1_1.0',
                                                       'md5hash': '12cd8e064eafadfc8f06af01d31fa8f5'},
                                                      {'version_name': 'raw_csv2_1.0',
                                                       'md5hash': 'f4466ed4b50c808bfeb64681b4f33dda'}]
        assert intake_source.metadata['md5hash'] == source.md5hash == '80d2fd6ad031412aa15c723dc550d81d'

        assert source.upstream_sources == [self.catalog['raw_csv1_1.0'], self.catalog['raw_csv2_1.0']]

        assert self.catalog.validate()[0]

    def test_validate(self):
        @analysis('raw_csv1', '1.0', catalog=self.catalog, sources=[], data_dir=self.data_dir)
        def load_raw_csv1(csv_source1):
            return csv_source1

        raw_csv1_source, _ = load_raw_csv1(self.csvsource1.to_dask())

        @analysis('transform_csv1', '1.0', catalog=self.catalog, sources=[raw_csv1_source], data_dir=self.data_dir)
        def transform_csv1(raw_csv1):
            raw_csv1['a'] *= 10
            return raw_csv1

        transform_csv1()

        source = self.catalog['transform_csv1_1.0']
        intake_source = source.intake_source
        assert intake_source.metadata['extension'] == 'framania'
        assert intake_source.metadata['version'] == source.version == '1.0'
        assert intake_source.metadata['version_name'] == 'transform_csv1_1.0'
        assert intake_source.metadata['name'] == source.name == 'transform_csv1'
        assert intake_source.metadata['upstream'] == [{'version_name': 'raw_csv1_1.0',
                                                       'md5hash': '12cd8e064eafadfc8f06af01d31fa8f5'}]
        assert intake_source.metadata['md5hash'] == source.md5hash == 'f4466ed4b50c808bfeb64681b4f33dda'

        assert source.upstream_sources == [self.catalog['raw_csv1_1.0']]

        raw_csv1_source, _ = load_raw_csv1(self.csvsource2.to_dask())
        assert not self.catalog.validate()[0]



from pathlib import Path
from typing import Union
from uuid import uuid4

import dask
import pandas
import yaml
from dask.dataframe import DataFrame
from intake import DataSource
from intake_parquet import ParquetSource


def initialize_catalog(catalog_file: Union[Path, str]):
    if isinstance(catalog_file, str):
        catalog_file = Path(catalog_file)
    try:
        catalog_contents = yaml.load(catalog_file.open().read(), Loader=yaml.SafeLoader)
    except FileNotFoundError:
        catalog_contents = {'sources': {}}
    if catalog_contents is None or \
            not isinstance(catalog_contents['sources'], dict) or \
            'sources' not in catalog_contents:
        catalog_contents = {'sources': {}}

    catalog_file.open('wt').write(yaml.dump(catalog_contents, Dumper=yaml.SafeDumper))


def add_source_to_catalog(source: DataSource, catalog_file: Union[Path, str]):
    """
    API to add new data source to catalog_file.

    Args:
        source: data source to add.
        catalog_file: file where data source to be added. if file doesn't exist, file will be created.
    Examples:
        >>> import os
        >>> from intake.source.csv import CSVSource
        >>> import yaml
        >>> source1 = CSVSource('test/temp/test1.csv')
        >>> source1.name = 'csv-test1'
        >>> # create new catalog with source
        >>> cfile = 'test/temp/test-catalog.yaml'
        >>> add_source_to_catalog(source1, cfile)
        >>> print(yaml.safe_load(Path(cfile).open().read()))
        ... # doctest: +NORMALIZE_WHITESPACE
        {'sources': {'csv-test1': {'args': {'urlpath': 'test/temp/test1.csv'}, 'description': '', 'driver': 'intake.source.csv.CSVSource', 'metadata': {}}}}
        >>> source2 = CSVSource('test/temp/test2.csv')
        >>> source2.name = 'csv-test2'
        >>> add_source_to_catalog(source2, cfile)
        >>> print(yaml.safe_load(Path(cfile).open().read()))
        ... # doctest: +NORMALIZE_WHITESPACE
        {'sources': {'csv-test1': {'args': {'urlpath': 'test/temp/test1.csv'}, 'description': '', 'driver': 'intake.source.csv.CSVSource', 'metadata': {}}, 'csv-test2': {'args': {'urlpath': 'test/temp/test2.csv'}, 'description': '', 'driver': 'intake.source.csv.CSVSource', 'metadata': {}}}}
        >>> os.remove(cfile)
    """
    if isinstance(catalog_file, str):
        catalog_file = Path(catalog_file)
    try:
        catalog_contents = yaml.load(catalog_file.open().read(), Loader=yaml.SafeLoader)
    except FileNotFoundError:
        catalog_contents = {'sources': {}}
    if catalog_contents is None or \
            not isinstance(catalog_contents['sources'], dict) or \
            'sources' not in catalog_contents:
        catalog_contents = {'sources': {}}
    catalog_contents['sources'][source.name] = source._yaml()['sources'][source.name]

    catalog_file.open('wt').write(yaml.dump(catalog_contents, Dumper=yaml.SafeDumper))


def dump_dask_to_intake(dd: DataFrame, data_name: str, data_dir: Union[str, Path], catalog_file: Union[Path, str],
                        **kwargs):
    """
    API to dump dask dataframe as parquet format and add it to intake catalog

    Args:
        dd: dask dataframe to dump.
        data_name: name used as a name of intake data source.
        data_dir: directory where dask dataframe will be stored.
        catalog_file: file where data source to be added. if file doesn't exist, file will be created.
        kwargs: Any options available for dask.dataframe.to_parquet. see https://docs.dask.org/en/latest/dataframe-api.html#dask.dataframe.to_parquet for detail.
    Returns:
        created parquet data source and dask to_parquet job (if you put compute=False in kwargs.)
    Examples:
        >>> import os
        >>> import shutil
        >>> import yaml
        >>> from intake.source.csv import CSVSource
        >>> import dask.dataframe
        >>> import pandas
        >>> pd = pandas.DataFrame({'a': [1, 2, 3, 4], 'b': [2, 3, 4, 5], 'c': [5, 6, 7, 8],
        ...                        'label': ['a', 'b', 'c', 'd']},
        ...                        index=[100, 200, 300, 400])
        >>> print(pd)
             a  b  c label
        100  1  2  5     a
        200  2  3  6     b
        300  3  4  7     c
        400  4  5  8     d
        >>> dd = dask.dataframe.from_pandas(pd, npartitions=2)
        >>> print(dd)
        ... # doctest: +NORMALIZE_WHITESPACE
        Dask DataFrame Structure:
                           a      b      c   label
        npartitions=2
        100            int64  int64  int64  object
        300              ...    ...    ...     ...
        400              ...    ...    ...     ...
        Dask Name: from_pandas, 2 tasks
        >>> cfile = 'test/temp/test-catalog.yaml'
        >>> ddir = 'test/temp/data-dir'
        >>> # DUMP WITH COMPUTATION
        >>> psource1, job = dump_dask_to_intake(dd, 'test-dd1', ddir, cfile)
        >>> print(psource1.name)
        test-dd1
        >>> print(psource1.read())
        ... # doctest: +NORMALIZE_WHITESPACE
               a  b  c label
        index
        100    1  2  5     a
        200    2  3  6     b
        300    3  4  7     c
        400    4  5  8     d
        >>> print(job is None)
        True
        >>> print(yaml.safe_load(Path(cfile).open().read()))
        ... # doctest: +NORMALIZE_WHITESPACE
        {'sources': {'test-dd1': {'args': {'urlpath': 'test/temp/data-dir/test-dd1'}, 'description': '', 'driver': 'intake_parquet.source.ParquetSource', 'metadata': {}}}}
        >>> # DUMP WITHOUT COMPUTATION
        >>> psource2, job = dump_dask_to_intake(dd, 'test-dd2', ddir, cfile, compute=False)
        >>> print(job is None)
        False
        >>> # do computation lazily
        >>> job.compute()
        >>> print(psource2.read())
        ... # doctest: +NORMALIZE_WHITESPACE
               a  b  c label
        index
        100    1  2  5     a
        200    2  3  6     b
        300    3  4  7     c
        400    4  5  8     d
        >>> os.remove(cfile)
        >>> shutil.rmtree(ddir)
    """
    if isinstance(data_dir, str):
        data_dir = Path(data_dir)

    parquet_dir = data_dir / data_name

    parquet_kwargs = {}
    if 'engine' in kwargs:
        parquet_kwargs['engine'] = kwargs['engine']

    psource = ParquetSource(str(parquet_dir), **parquet_kwargs)
    psource.name = data_name
    add_source_to_catalog(psource, catalog_file)
    to_parquet_result = dd.to_parquet(str(parquet_dir), **kwargs)

    return psource, to_parquet_result


def dump_pandas_to_intake(pd: pandas.DataFrame, data_name: str, data_dir: Union[str, Path],
                          catalog_file: Union[Path, str], npartitions: int = 1, **kwargs):
    """
    API to dump pandas dataframe as parquet format and add it to intake catalog.
    pandas dataframe will be converted to dask dataframe automatically internally.

    Args:
        pd: pandas dataframe to dump.
        data_name: name used as a name of intake data source.
        data_dir: directory where dask dataframe will be stored.
        catalog_file: file where data source to be added. if file doesn't exist, file will be created.
        npartitions: number of partitions input pandas dataframe split into.
        kwargs: Any options available for dask.dataframe.to_parquet. see https://docs.dask.org/en/latest/dataframe-api.html#dask.dataframe.to_parquet for detail.
    Returns:
        created parquet data source and dask to_parquet job (if you put compute=False in kwargs.)
    Examples:
        >>> import os
        >>> import shutil
        >>> import yaml
        >>> from intake.source.csv import CSVSource
        >>> import dask.dataframe
        >>> import pandas
        >>> pd = pandas.DataFrame({'a': [1, 2, 3, 4], 'b': [2, 3, 4, 5], 'c': [5, 6, 7, 8],
        ...                        'label': ['a', 'b', 'c', 'd']},
        ...                        index=[100, 200, 300, 400])
        >>> print(pd)
             a  b  c label
        100  1  2  5     a
        200  2  3  6     b
        300  3  4  7     c
        400  4  5  8     d
        >>> cfile = 'test/temp/test-catalog.yaml'
        >>> ddir = 'test/temp/data-dir'
        >>> # DUMP WITH COMPUTATION
        >>> psource1, job = dump_pandas_to_intake(pd, 'test-pd1', ddir, cfile)
        >>> print(psource1.name)
        test-pd1
        >>> print(psource1.read())
        ... # doctest: +NORMALIZE_WHITESPACE
               a  b  c label
        index
        100    1  2  5     a
        200    2  3  6     b
        300    3  4  7     c
        400    4  5  8     d
        >>> print(job is None)
        True
        >>> print(yaml.safe_load(Path(cfile).open().read()))
        ... # doctest: +NORMALIZE_WHITESPACE
        {'sources': {'test-pd1': {'args': {'urlpath': 'test/temp/data-dir/test-pd1'}, 'description': '', 'driver': 'intake_parquet.source.ParquetSource', 'metadata': {}}}}
        >>> # DUMP WITHOUT COMPUTATION
        >>> psource2, job = dump_pandas_to_intake(pd, 'test-pd2', ddir, cfile, compute=False)
        >>> print(job is None)
        False
        >>> # do computation lazily
        >>> job.compute()
        >>> print(psource2.read())
        ... # doctest: +NORMALIZE_WHITESPACE
               a  b  c label
        index
        100    1  2  5     a
        200    2  3  6     b
        300    3  4  7     c
        400    4  5  8     d
        >>> os.remove(cfile)
        >>> shutil.rmtree(ddir)
    """
    dd = dask.dataframe.from_pandas(pd, npartitions=npartitions)
    return dump_dask_to_intake(dd, data_name, data_dir, catalog_file, **kwargs)


def persist_local(dd: DataFrame, persist_dir: Union[str, Path], **kwargs):
    """
    API to persist dask dataframe as perquet files on disk.

    Args:
        dd: dask dataframe to persist.
        persist_dir: directory where dask dataframe will be stored.
        kwargs: Any options available for dask.dataframe.to_parquet. see https://docs.dask.org/en/latest/dataframe-api.html#dask.dataframe.to_parquet for detail.
    Returns:
        created parquet data source and dask to_parquet job (if you put compute=False in kwargs.)
    Examples:
        >>> import os
        >>> import shutil
        >>> import yaml
        >>> from intake.source.csv import CSVSource
        >>> import dask.dataframe
        >>> import pandas
        >>> pd = pandas.DataFrame({'a': [1, 2, 3, 4], 'b': [2, 3, 4, 5], 'c': [5, 6, 7, 8],
        ...                        'label': ['a', 'b', 'c', 'd']},
        ...                        index=[100, 200, 300, 400])
        >>> print(pd)
        ... # doctest: +NORMALIZE_WHITESPACE
             a  b  c label
        100  1  2  5     a
        200  2  3  6     b
        300  3  4  7     c
        400  4  5  8     d
        >>> dd = dask.dataframe.from_pandas(pd, npartitions=2)
        >>> ddir = 'test/temp/data-dir/persist'
        >>> dd2 = persist_local(dd, ddir)
        >>> persist_id = dd2.persist_id
        >>> print(dd2.compute())
        ... # doctest: +NORMALIZE_WHITESPACE
               a  b  c label
        index
        100    1  2  5     a
        200    2  3  6     b
        300    3  4  7     c
        400    4  5  8     d
        >>> assert (Path(ddir) / persist_id).exists()
        >>> shutil.rmtree(ddir)
    """
    data_id = str(uuid4())
    persist_dir = persist_dir if isinstance(persist_dir, Path) else Path(persist_dir)

    persist_dir.mkdir(parents=True, exist_ok=True)

    persist_catalog_file = Path(persist_dir) / 'persist_catalog.yaml'
    psource, _ = dump_dask_to_intake(dd, data_id, persist_dir, persist_catalog_file, **kwargs)

    new_dd = psource.to_dask()
    new_dd.persist_id = data_id

    return new_dd



import pandas
from framania.pandasmania.transform import grouper_for_timeseries, group_row_number


def fbfill_series(ps: pandas.Series, limit=None):
    """
        API to fill na values by closer one of last value or next value. (closer means one of row position is close.)
        NOTICE Position of row will NOT be sorted by index automatically. Original order of rows is kept.

        Args:
            ps (pandas.Series): target pandas dataframe
            limit (optional int): see https://pandas.pydata.org/pandas-docs/version/0.23/generated/pandas.DataFrame.fillna.html#pandas.DataFrame.fillna
        Returns:
            new pandas series with filled values
        Examples:
            >>> import pandas
            >>> pd = pandas.DataFrame({'a': [1, 2, 3, None, None, None, 4, 5, None, 6, None, None, 7, None, 8]},
            ...                       index=[0, 14, 1, 13, 2, 12, 3, 11, 4, 10, 5, 9, 6, 8, 7])
            >>> print(pd)
            ... # doctest: +NORMALIZE_WHITESPACE
                  a
            0   1.0
            14  2.0
            1   3.0
            13  NaN
            2   NaN
            12  NaN
            3   4.0
            11  5.0
            4   NaN
            10  6.0
            5   NaN
            9   NaN
            6   7.0
            8   NaN
            7   8.0
            >>> result = fbfill_series(pd.a)
            >>> print(result)
            ... # doctest: +NORMALIZE_WHITESPACE
            0     1.0
            14    2.0
            1     3.0
            13    3.0
            2     3.0
            12    4.0
            3     4.0
            11    5.0
            4     5.0
            10    6.0
            5     6.0
            9     7.0
            6     7.0
            8     7.0
            7     8.0
            Name: a, dtype: float64
        """
    ffill_series = ps.fillna(method='ffill', axis=0, inplace=False, limit=limit)
    bfill_series = ps.fillna(method='bfill', axis=0, inplace=False, limit=limit)

    nan_value_flags = ps.isnull() & ~ps.shift(1).isnull()
    temp_df = pandas.DataFrame({'v': ps, 'na_flag': nan_value_flags}, index=ps.index)
    temp_df['nan_group'] = grouper_for_timeseries(temp_df.na_flag)
    temp_df = temp_df[ps.isnull()]
    temp_df['fnrow'] = group_row_number(temp_df, ['nan_group'], True)
    temp_df['bnrow'] = group_row_number(temp_df, ['nan_group'], False)

    ffill_index = ps.isnull() & (temp_df.fnrow <= temp_df.bnrow)
    rfill_index = ps.isnull() & (temp_df.fnrow > temp_df.bnrow)

    result = pandas.Series(ps)
    result.loc[ffill_index] = ffill_series.loc[ffill_index]
    result.loc[rfill_index] = bfill_series.loc[rfill_index]
    return result



from typing import List, Any

import numpy
import pandas
from pandas import Series, DataFrame


def timeseries_value_changed(value_series: Series, abs_threshold: float=0.0, change_at_nan: bool = True):
    """
    API to check if value of series in sorted timeseries data is changed.

    Args:
        value_series (pandas.Series): series contains some value
        abs_threshold (float): absolute threshold to check value is changed or not. default: 0.0 (any changes)
        change_at_nan (bool): if True, consider that nan means value is changed
    Returns:
        result bool pandas series
    Examples:
        >>> import pandas
        >>> pd = pandas.DataFrame({
        ...     'a': [1, 2, None, 8, 20, 23, 28, 30]
        ... })
        >>> print(pd)
        ... # doctest: +NORMALIZE_WHITESPACE
             a
        0   1.0
        1   2.0
        2   NaN
        3   8.0
        4  20.0
        5  23.0
        6  28.0
        7  30.0
        >>> result = timeseries_value_changed(pd.a, 10)
        >>> print(result)
        0     True
        1    False
        2     True
        3     True
        4     True
        5    False
        6    False
        7    False
        Name: a, dtype: bool
        >>> result = timeseries_value_changed(pd.a, 10, change_at_nan=False)
        >>> print(result)
        0     True
        1    False
        2    False
        3    False
        4     True
        5    False
        6    False
        7    False
        Name: a, dtype: bool
    """


    if change_at_nan is False:
        value_series = value_series.fillna(method='ffill')

    last_value_series = value_series.shift(1)
    value_changed = ~((value_series - last_value_series).abs() < abs_threshold)
    return value_changed


def timeseries_id_changed(id_series: Series, change_at_nan: bool = True):
    """
    API to check if value of id series in sorted timeseries data is changed. (sort key must be `[id, time]`)

    Args:
        id_series (pandas.Series): series contains some id
        change_at_nan (bool): if True, consider that nan means id is changed
    Returns:
        result bool pandas series
    Examples:
        >>> import pandas
        >>> pd = pandas.DataFrame({
        ...     'a': [1, 1, 1, 1, 2, None, 3, 3]
        ... })
        >>> print(pd)
        ... # doctest: +NORMALIZE_WHITESPACE
             a
        0  1.0
        1  1.0
        2  1.0
        3  1.0
        4  2.0
        5  NaN
        6  3.0
        7  3.0
        >>> result = timeseries_id_changed(pd.a)
        >>> print(result)
        0     True
        1    False
        2    False
        3    False
        4     True
        5     True
        6     True
        7    False
        Name: a, dtype: bool
        >>> result = timeseries_id_changed(pd.a, change_at_nan=False)
        >>> print(result)
        0     True
        1    False
        2    False
        3    False
        4     True
        5    False
        6     True
        7    False
        Name: a, dtype: bool
    """


    if change_at_nan is False:
        id_series = id_series.fillna(method='ffill')

    last_value_series = id_series.shift(1)

    value_changed = id_series != last_value_series
    return value_changed



def grouper_for_timeseries(split_flags: Series, start_id: int = 1):
    """
    API to build grouper id to split timeseries data by some splitting rules.

    Args:
        split_flags (pandas.Series[bool]): bool type pandas series, with True value at row where to be split
        start_id (int): start value for incremental grouper id
    Returns:
        result pandas series
    Examples:
        >>> import pandas
        >>> pd = pandas.DataFrame({
        ...     'a': [1, 2, 5, 8, 20, 23, 28, 30],
        ...     'b': [1, 1, 1, 1, 2, 2, 3, 3]
        ... })
        >>> print(pd)
        ... # doctest: +NORMALIZE_WHITESPACE
            a  b
        0   1  1
        1   2  1
        2   5  1
        3   8  1
        4  20  2
        5  23  2
        6  28  3
        7  30  3
        >>> splitter = timeseries_id_changed(pd.b) | timeseries_value_changed(pd.a, 10)
        >>> print(splitter)
        0     True
        1    False
        2    False
        3    False
        4     True
        5    False
        6     True
        7    False
        dtype: bool
        >>> result = grouper_for_timeseries(splitter)
        >>> print(result)
        ... # doctest: +NORMALIZE_WHITESPACE
        0    1
        1    1
        2    1
        3    1
        4    2
        5    2
        6    3
        7    3
        dtype: int64
    """

    return split_flags.astype(int).cumsum() + (start_id - 1)


def group_row_number(pd: DataFrame, groupby: List[Any], ascending: bool=True):
    """
        API to add series that contains numbers of row in each groupby group.

        Args:
            pd (pandas.DataFrame): pandas dataframe which contains groupby columns
            groupby (List[Any]): grouping key. row number will be calculated in each of this groupby
            ascending (bool): if True, row number will starts with 0 and ends with (number of rows - 1). False means opposite
        Returns:
            result row number pandas series
        Examples:
            >>> import pandas
            >>> pd = pandas.DataFrame({
            ...     'a': [1, 1, 1, 1, 2, 2, 2, 2, 3, 3, None, 4],
            ...     'b': [100, 100, 200, 300, 50, 60, 70, 80, 10, 20, -1, 100]
            ... })
            >>> print(pd)
            ... # doctest: +NORMALIZE_WHITESPACE
                  a    b
            0   1.0  100
            1   1.0  100
            2   1.0  200
            3   1.0  300
            4   2.0   50
            5   2.0   60
            6   2.0   70
            7   2.0   80
            8   3.0   10
            9   3.0   20
            10  NaN   -1
            11  4.0  100
            >>> result = group_row_number(pd, ['a'])
            >>> print(result)
            0     0
            1     1
            2     2
            3     3
            4     0
            5     1
            6     2
            7     3
            8     0
            9     1
            10    0
            11    0
            dtype: int64
            >>> result = group_row_number(pd, ['a'], ascending=False)
            >>> print(result)
            0     3
            1     2
            2     1
            3     0
            4     3
            5     2
            6     1
            7     0
            8     1
            9     0
            10    0
            11    0
            Name: row_number, dtype: int64

        """
    temp_df = pd[groupby]
    g = temp_df.groupby(groupby, as_index=False)
    row_number = g.cumcount()
    if ascending:
        return row_number

    temp_df.loc[:, 'row_number'] = row_number
    row_count = g.count()
    m_temp_df = temp_df.merge(row_count, on=groupby, how='left')
    m_temp_df.index = temp_df.index

    m_temp_df['row_number'] = \
        m_temp_df['row_number_y'].fillna(1).astype(int) - m_temp_df['row_number_x'] - 1
    return m_temp_df['row_number']

def asstr(ps: pandas.Series, keep_nan: bool = True, remove_dotzero: bool = False):
    """
    API to make pandas Series as type string keeping NaN value.

    Args:
        ps (pandas.Series): target pandas series
        keep_nan (bool): if keep NaN as NaN or make nan as string "nan"
        remove_dotzero (bool): if remove
    Returns:
        string type pandas Series
    Examples:
        >>> import pandas
        >>> pd = pandas.DataFrame({
        ...     'a': [1, 1, 1, 1, 2, 2, 2, 2, 3, 3, None, 4.01],
        ...     'b': [100, 100, 200, 300, 50, 60, 70, 80, 10, 20, -1, 100]
        ... })
        >>> print(pd)
        ... # doctest: +NORMALIZE_WHITESPACE
               a    b
        0   1.00  100
        1   1.00  100
        2   1.00  200
        3   1.00  300
        4   2.00   50
        5   2.00   60
        6   2.00   70
        7   2.00   80
        8   3.00   10
        9   3.00   20
        10   NaN   -1
        11  4.01  100
        >>> result = asstr(pd.a)
        >>> print(result)
        0      1.0
        1      1.0
        2      1.0
        3      1.0
        4      2.0
        5      2.0
        6      2.0
        7      2.0
        8      3.0
        9      3.0
        10     NaN
        11    4.01
        Name: a, dtype: object
        >>> result = asstr(pd.a, remove_dotzero=True)
        >>> print(result)
        0        1
        1        1
        2        1
        3        1
        4        2
        5        2
        6        2
        7        2
        8        3
        9        3
        10     NaN
        11    4.01
        Name: a, dtype: object
    """

    if not keep_nan:
        return ps.astype('str')

    str_s: pandas.Series = ps.astype('str')
    str_s[ps.isnull()] = numpy.nan

    if remove_dotzero:
        str_s = str_s.str.replace('.0$', '')

    return str_s



from typing import Any

import pandas
from pandas import DataFrame, MultiIndex
from pandas.core.groupby import GroupBy


def flatten_aggregated_columns(pd: DataFrame):
    """
    API to make aggregated columns that are MultiIndex style flat

    Args:
        pd: target dataframe
    Returns:
        result pandas dataframe
    Examples:
        >>> import dask.dataframe
        >>> import pandas
        >>> pd = pandas.DataFrame({'a': [1, 10, 100, 1, 1, 100], 'b': range(0, 600, 100), 'key': [0, 1, 2, 0, 1, 2]})
        >>> print(pd)
        ... # doctest: +NORMALIZE_WHITESPACE
             a    b  key
        0    1    0    0
        1   10  100    1
        2  100  200    2
        3    1  300    0
        4    1  400    1
        5  100  500    2
        >>> groupby = pd.groupby(['key'])
        >>> groupby_result = groupby.agg({'a': ['sum', 'min'], 'b': ['mean', 'sum', 'max']})
        >>> print(groupby_result)
        ... # doctest: +NORMALIZE_WHITESPACE
               a         b
             sum  min mean  sum  max
        key
        0      2    1  150  300  300
        1     11    1  250  500  400
        2    200  100  350  700  500
        >>> flatten = flatten_aggregated_columns(groupby_result)
        >>> print(flatten)
        ... # doctest: +NORMALIZE_WHITESPACE
             a_sum  a_min  b_mean  b_sum  b_max
        key
        0        2      1     150    300    300
        1       11      1     250    500    400
        2      200    100     350    700    500
        >>> print(flatten_aggregated_columns(pd))
        ... # doctest: +NORMALIZE_WHITESPACE
             a    b  key
        0    1    0    0
        1   10  100    1
        2  100  200    2
        3    1  300    0
        4    1  400    1
        5  100  500    2
    """
    if not isinstance(pd.columns, MultiIndex) or pd.columns.nlevels != 2:
        return pd

    columns = []
    for l1, l2 in zip(pd.columns.get_level_values(0), pd.columns.get_level_values(1)):
        if l2 == '':
            columns.append(l1)
        else:
            columns.append('_'.join((l1, l2)))
    result = pd.set_axis(columns, axis=1, inplace=False)
    return result

def aggregate_as_series(groupby: GroupBy, series: Any, agg_func: Any):
    """
    API to output the aggregate result as a column of the original data frame.

    Args:
        groupby (GroupBy): result of pandas.DataFrame.groupby
        series (Any): name of series to be aggregated
        agg_func (Union[str, List[str]]): see `func` in https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.agg.html
    Returns:
        result pandas series
    Examples:
        >>> import dask.dataframe
        >>> import pandas
        >>> pd = pandas.DataFrame({'a': [1, 10, 100, 1, 1, 100], 'b': range(0, 600, 100), 'key': [0, 1, 2, 0, 1, 2]})
        >>> print(pd)
        ... # doctest: +NORMALIZE_WHITESPACE
             a    b  key
        0    1    0    0
        1   10  100    1
        2  100  200    2
        3    1  300    0
        4    1  400    1
        5  100  500    2
        >>> groupby = pd.groupby(['key', 'a'])
        >>> pd['sum_b'] = aggregate_as_series(groupby, 'b', 'sum')
        >>> print(pd)
        ... # doctest: +NORMALIZE_WHITESPACE
             a    b  key  sum_b
        0    1    0    0    300
        1   10  100    1    100
        2  100  200    2    700
        3    1  300    0    300
        4    1  400    1    400
        5  100  500    2    700
    """
    def _series_agg(s, agg_func):
        new_s = pandas.Series(s.agg(agg_func), index=s.index)
        return new_s

    return groupby[series].apply(_series_agg, agg_func=agg_func)



from collections import OrderedDict
from typing import Any, List, Tuple, Callable

import pandas
from dask.dataframe import DataFrame


def make_meta(index_name_type: Tuple[Any, Any], column_name_types: List[Tuple[Any, Any]]) -> pandas.DataFrame:
    """
    API to make dask _meta from index schema info and column schema info.

    Args:
            index_name_type (Tuple[Any, Any): Tuple like (index_name, index_type). index_name can be None.
            column_name_types (List[Tuple[Any, Any]]): List of Tuple like (column name, column type). The order of it will be kept in meta.
        Returns:
            result meta as an empty pandas DataFrame
        Examples:
            >>> result = make_meta(('idx', 'object'),
            ...                    [('int_value', 'int'), ('str_value', 'object'), ('float_value', 'float')])
            >>> print(result.columns)
            Index(['int_value', 'str_value', 'float_value'], dtype='object')
            >>> print(result.dtypes)
            int_value        int64
            str_value       object
            float_value    float64
            dtype: object
            >>> print(result.index)
            Index([], dtype='object', name='idx')
    """
    index_name, index_type = index_name_type
    idx = pandas.Index(data=[], name=index_name, dtype=index_type)
    columns = OrderedDict([(c_name, pandas.Series(data=[], index=idx, dtype=c_type, name=c_name))
                           for c_name, c_type in column_name_types])
    return pandas.DataFrame(columns, index=idx)


def map_partitions_as_meta(dd: DataFrame, func: Callable[..., pandas.DataFrame],
                           meta: pandas.DataFrame, **kwargs):
    """
    API to do map_partitions, and reformat result using meta. It may avoid error caused by map_partitions result and meta don't match.

    Args:
            dd (DataFrame): dask dataframe to do map_partitions.
            func (Callable[[[pandas.DataFrame, ...]], pandas.DataFrame]): function for map_partitions.
            meta (pandas.DataFrame): expected schema of map_partitions result.
            kwargs: additional arguments for func.
        Returns:
            result dask dataframe
        Examples:
            >>> import dask.dataframe
            >>> import pandas
            >>> pd = pandas.DataFrame({'a1': ['1,2,3', '2,3,4', '3,4'], 'a2': ['a,b,c', 'b,c,d', 'c,d'],
            ...                        'b': [1, 2, 3], 'idx': [0, 1, 2]}).set_index('idx')
            >>> print(pd)
            ... # doctest: +NORMALIZE_WHITESPACE
                    a1     a2  b
            idx
            0    1,2,3  a,b,c  1
            1    2,3,4  b,c,d  2
            2      3,4    c,d  3
            >>> transformer = lambda pd: pd[['b', 'a1', 'a2']]
            >>> dd = dask.dataframe.from_pandas(pd, npartitions=2)
            >>> result1 = dd.map_partitions(transformer)
            >>> print(result1.compute())
            ... # doctest: +NORMALIZE_WHITESPACE
                 b     a1     a2
            idx
            0    1  1,2,3  a,b,c
            1    2  2,3,4  b,c,d
            2    3    3,4    c,d
            >>> meta = make_meta(('idx', 'int'), [('a1', 'object'), ('a2', 'object'), ('b', 'int')])
            >>> result2 = map_partitions_as_meta(dd, transformer, meta)
            >>> print(result2.compute())
            ... # doctest: +NORMALIZE_WHITESPACE
                    a1     a2  b
            idx
            0    1,2,3  a,b,c  1
            1    2,3,4  b,c,d  2
            2      3,4    c,d  3
            >>> transformer2 = lambda pd, v: pd[['b']] * v
            >>> result3 = map_partitions_as_meta(dd, transformer2, make_meta(('idx', 'int'), [('b', 'int')]),
            ...                                  v=100)
            >>> print(result3.compute())
            ... # doctest: +NORMALIZE_WHITESPACE
                   b
            idx
            0    100
            1    200
            2    300
    """

    def apply_meta(pd: pandas.DataFrame):
        result = func(pd, **kwargs)
        meta_column_names = [c for c in meta.columns]
        return result[meta_column_names]

    return dd.map_partitions(apply_meta)



import re
from pathlib import Path
from typing import List, Optional, Any, Union, Dict, Tuple

import pandas
from dask.dataframe import DataFrame, from_pandas
from intake import DataSource, Catalog
from intake_parquet import ParquetSource

from framania.daskmania.util import md5hash as ddmd5hash
from framania.intakemania.util import add_source_to_catalog, initialize_catalog


def parse_version(v: Any):
    if not isinstance(v, str):
        return v

    version_info = tuple(part if not part.isnumeric() else int(part) for part in v.split('.'))
    return version_info


def get_version_name(name: str, version: str):
    return f'{name}_{version}'


def version_and_name(version_name: str):
    if '_' not in version_name:
        raise Exception(f'Invalid version_name {version_name}.')

    split = version_name.rsplit('_', maxsplit=1)
    return split[0], split[1]


def validate_version(version: str):
    version.replace('_', '.')


class FramaniaExtendedIntakeCatalog:
    def __init__(self, intake_catalog_file: Path):
        try:
            self.intake_catalog = Catalog(str(intake_catalog_file))
        except:
            initialize_catalog(intake_catalog_file)
            self.intake_catalog = Catalog(str(intake_catalog_file))

    def find_by_version_name(self, version_name: str) -> 'FramaniaExtendedIntakeSource':
        source = self.intake_catalog[version_name]

        source_extension = source.metadata.get('extension', None)
        if source_extension != 'framania':
            raise Exception(f'Data source {version_name} is not framania extended.')

        name, version = version_and_name(version_name)
        assert version == source.metadata['version']
        md5hash = source.metadata['md5hash']

        upstream_sources = [self.find_validated_by_version_name(s['version_name'], s['md5hash']) for s in
                            source.metadata['upstream']]
        return FramaniaExtendedIntakeSource(source, name, version, md5hash, upstream_sources, existing_source=True)

    def find_by_version_and_name(self, name: str, version: str) -> 'FramaniaExtendedIntakeSource':
        return self.find_by_version_name(get_version_name(name, version))

    def find_validated_by_version_name(self, version_name: str, md5hash: str) -> 'FramaniaExtendedIntakeSource':
        found = self.find_by_version_name(version_name)
        assert found.md5hash == md5hash
        return found

    def find_latest_source_by_name(self, name: str) -> 'FramaniaExtendedIntakeSource':
        sources = self.intake_catalog._entries
        source_names = [k for k in sources.keys() if k.startswith(name)]

        versions = {}

        for source_name in source_names:
            source: DataSource = self.intake_catalog[source_name]
            source_version = source.metadata['version']

            if source_name != get_version_name(name, source_version):
                continue

            versions[source_version] = source_name

        latest_version = list(sorted(versions.keys(), key=lambda v: parse_version(v), reverse=True))[0]
        return self.find_by_version_name(versions[latest_version])

    def __getitem__(self, item: str) -> 'FramaniaExtendedIntakeSource':
        try:
            return self.find_by_version_name(item)
        except:
            return self.find_latest_source_by_name(item)

    def append(self, value: 'FramaniaExtendedIntakeSource'):
        add_source_to_catalog(value.intake_source, self.intake_catalog.path)
        self.intake_catalog = Catalog(self.intake_catalog.path)

    def validate(self) -> Tuple[bool, Dict]:
        result = {}
        result_flag = True

        all_entries = list(self.intake_catalog._entries.keys())

        finished = set()
        l_entries = len(all_entries)
        for i in range(l_entries):
            for j in range(l_entries):
                entry = all_entries[j]

                if entry in finished:
                    continue

                source = self.intake_catalog[entry]
                if source.metadata.get('extension', '') != 'framania':
                    continue

                if any([s['version_name'] not in finished for s in source.metadata['upstream']]):
                    continue

                result[entry] = []

                validate_flag = True
                name, version = version_and_name(entry)
                validate_flag = validate_flag and (source.metadata.get('name', '') == name)
                validate_flag = validate_flag and (source.metadata.get('version', '') == version)
                validate_flag = validate_flag and (source.metadata.get('version_name', '') == entry)

                for s in source.metadata['upstream']:
                    try:
                        self.find_validated_by_version_name(s['version_name'], s['md5hash'])
                        flag = result[s['version_name']][-1][1]
                    except AssertionError:
                        flag = False
                    validate_flag = validate_flag and flag
                    result[entry].append((s['version_name'], flag))

                result[entry].append((entry, validate_flag))
                result_flag = result_flag and validate_flag
                finished.add(entry)
        return result_flag, result

    def dump_dask(self, dd: DataFrame, data_name: str, version: str,
                  data_dir: Union[str, Path], upstream_sources: List['FramaniaExtendedIntakeSource'] = None,
                  **kwargs):
        if isinstance(data_dir, str):
            data_dir = Path(data_dir)

        parquet_dir = data_dir / data_name / version

        parquet_kwargs = {}
        if 'engine' in kwargs:
            parquet_kwargs['engine'] = kwargs['engine']

        to_parquet_result = dd.to_parquet(str(parquet_dir), **kwargs)

        psource = ParquetSource(str(parquet_dir), **parquet_kwargs)
        md5hash = ddmd5hash(psource.to_dask())
        framania_psource = FramaniaExtendedIntakeSource(psource, data_name, version,
                                                        md5hash, upstream_sources)
        self.append(framania_psource)

        return framania_psource, to_parquet_result

    def dump_pandas(self, pd: pandas.DataFrame, data_name: str, version: str,
                    data_dir: Union[str, Path], upstream_sources: List['FramaniaExtendedIntakeSource'] = None,
                    npartitions: int = 1, **kwargs):
        dd = from_pandas(pd, npartitions=npartitions)
        self.dump_dask(dd, data_name, version, data_dir, upstream_sources, **kwargs)


class FramaniaExtendedIntakeSource:
    good_name_chars = '[0-9|a-z|A-Z|\-|_]+'
    good_version_chars = '[0-9|a-z|A-Z|\-|.]+'

    def __init__(self, intake_source: DataSource,
                 name: str, version: str,
                 md5hash: Optional[str] = None,
                 upstream_sources: Optional[List['FramaniaExtendedIntakeSource']] = None,
                 existing_source: bool = False,
                 metadata: Dict[Any, Any] = None):

        if re.match(self.good_name_chars, name) is None:
            raise Exception('Valid data name characters: [0-9, a-z, A-Z, -, _].')
        if re.match(self.good_version_chars, version) is None:
            raise Exception('Valid data version characters: [0-9, a-z, A-Z, -, .].')

        if metadata is None:
            metadata = {}

        self.intake_source = intake_source
        self.name = name
        self.version = version
        self.md5hash = md5hash
        self.upstream_sources = upstream_sources or []

        if not existing_source:
            vname = get_version_name(self.name, self.version)
            self.intake_source.name = vname
            self.intake_source.metadata['version'] = self.version
            self.intake_source.metadata['extension'] = 'framania'
            self.intake_source.metadata['name'] = self.name
            self.intake_source.metadata['version_name'] = vname
            self.intake_source.metadata['md5hash'] = self.md5hash

            upstream_sources = [{'version_name': get_version_name(source.name, source.version),
                                 'md5hash': source.md5hash}
                                for source in self.upstream_sources]

            self.intake_source.metadata['upstream'] = upstream_sources
            self.intake_source.metadata.update(metadata)

    def update_md5hash(self, md5hash):
        self.md5hash = md5hash
        self.intake_source.metadata['md5hash'] = md5hash

    def __eq__(self, other):
        return (self.name == other.name) and \
               (self.version == other.version) and \
               (self.md5hash == other.md5hash) and \
               (self.upstream_sources == other.upstream_sources)


def analysis(name: str, version: str,
             catalog: FramaniaExtendedIntakeCatalog, sources: List[str],
             data_dir: Path):
    """
    Examples:
        >>> import pandas
        >>> import dask.dataframe
        >>> from pathlib import Path
        >>> import os
        >>> import shutil
        >>> ddir = Path('test/temp/data-dir')
        >>> cfile = Path('test/temp/test-catalog.yaml')
        >>> catalog = FramaniaExtendedIntakeCatalog(cfile)
        >>> @analysis('test1', '1.0', catalog, [], ddir)
        ... def test1():
        ...     df = pandas.DataFrame({'a': [1,2,3], 'b': [2,3,4]})
        ...     ddf = dask.dataframe.from_pandas(df, npartitions=1)
        ...     return ddf
        >>> @analysis('test2', '1.0', catalog, ['test1'], ddir)
        ... def test2(test1):
        ...     test1['c'] = test1.a * test1.b
        ...     return test1
        >>> _ = test1()
        >>> _ = test2()
        >>> os.remove(str(cfile))
        >>> shutil.rmtree(str(ddir))
    """

    def decorator_analysis(func):
        def wrapper(*args, **kwargs) -> Tuple[FramaniaExtendedIntakeSource, Any]:
            upstream_sources = [catalog[source] for source in sources]
            input: Dict[str, FramaniaExtendedIntakeSource] = \
                {source.name: source.intake_source.to_dask() for source in upstream_sources}
            result_dask = func(*args, **kwargs, **input)
            result_source, dask_job = catalog.dump_dask(result_dask, name, version, data_dir, upstream_sources)
            return result_source, dask_job

        return wrapper

    return decorator_analysis
